# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from qrl.core import logger
from qrl.core.Block import Block

logger.initialize_default()


class TestBlock(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestBlock, self).__init__(*args, **kwargs)

    def test_init(self):
        # TODO: Not much going on here..
        block = Block()
        self.assertIsNotNone(block)             # just to avoid warnings



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from mock import Mock, MagicMock

from qrl.core import logger
from qrl.core.ESyncState import ESyncState
from qrl.core.GenesisBlock import GenesisBlock
from qrl.core.node import POS

logger.initialize_default()


class TestNode(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestNode, self).__init__(*args, **kwargs)

    def test_create(self):
        buffered_chain = Mock()
        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POS(buffered_chain=buffered_chain,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider)

        self.assertIsNotNone(node)

    def test_sync_state_change_unsynced(self):
        buffered_chain = Mock()
        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POS(buffered_chain=buffered_chain,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider)

        self.assertIsNotNone(node)
        node.update_node_state(ESyncState.unsynced)
        # FIXME: Add more asserts

    def test_sync_state_change_syncing(self):
        buffered_chain = Mock()
        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POS(buffered_chain=buffered_chain,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider)

        self.assertIsNotNone(node)
        node.update_node_state(ESyncState.syncing)
        # FIXME: Add more asserts

    def test_sync_state_change_synced(self):
        buffered_chain = Mock()
        buffered_chain.height = 0
        buffered_chain.get_block = MagicMock(return_value=GenesisBlock())

        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POS(buffered_chain=buffered_chain,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider)

        self.assertIsNotNone(node)
        node.update_node_state(ESyncState.synced)
        # FIXME: Add more asserts

    def test_sync_state_change_forked(self):
        buffered_chain = Mock()
        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POS(buffered_chain=buffered_chain,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider)

        self.assertIsNotNone(node)
        node.update_node_state(ESyncState.forked)
        # FIXME: Add more asserts



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from google.protobuf.json_format import MessageToJson, Parse

from qrl.core import config
from qrl.core.Transaction import StakeTransaction
from qrl.generated import qrl_pb2
from qrl.crypto.misc import sha256_n


class StakeValidator:
    """
    Stake Validator class to represent each unique Stake Validator

    Maintains the cache of successfully validated hashes, saves validation
    time by avoiding recalculation of the hash till the hash terminators.
    """

    def __init__(self, stakevalidator_protobuf=None):
        self._data = stakevalidator_protobuf
        if not self._data:
            self._data = qrl_pb2.StakeValidator()

    @property
    def pbdata(self):
        return self._data

    @property
    def address(self) -> bytes:
        return self._data.address

    @property
    def slave_public_key(self) -> bytes:
        return self._data.slave_public_key

    @property
    def terminator_hash(self) -> bytes:
        return self._data.terminator_hash

    @property
    def balance(self) -> int:
        return self._data.balance

    @property
    def is_banned(self) -> bool:
        return self._data.is_banned

    @property
    def is_active(self) -> bool:
        return self._data.is_active

    @property
    def nonce(self) -> int:
        return self._data.nonce

    @property
    def activation_blocknumber(self) -> int:
        return self._data.activation_blocknumber

    def increase_nonce(self):
        self._data.nonce += 1

    @staticmethod
    def _hash_to_terminator(reveal_hash: bytes, times: int) -> bytes:
        return sha256_n(reveal_hash, times)

    @staticmethod
    def create(balance: int,
               stake_txn: StakeTransaction):

        stakevalidator = StakeValidator()

        stakevalidator._data.address = stake_txn.txfrom
        stakevalidator._data.slave_public_key = stake_txn.slave_public_key

        stakevalidator._data.terminator_hash = stake_txn.hash
        if not stakevalidator._data.terminator_hash:
            raise ValueError("terminator hash cannot be empty")

        stakevalidator._data.balance = balance
        if balance < config.dev.minimum_staking_balance_required:
            raise ValueError("balance should be at least {}".format(config.dev.minimum_staking_balance_required))

        stakevalidator._data.activation_blocknumber = stake_txn.activation_blocknumber

        stakevalidator._data.nonce = 0
        stakevalidator._data.is_banned = False
        stakevalidator._data.is_active = True  # Flag that represents if the stakevalidator has been deactivated by destake txn

        return stakevalidator

    def validate_hash(self, reveal_hash: bytes, block_idx: int) -> bool:
        # FIXME: Measure with a profiler if we really need a cache here

        times = block_idx - self.activation_blocknumber + 1
        terminator_found = self._hash_to_terminator(reveal_hash, times)
        terminator_expected = self.terminator_hash

        return terminator_found == terminator_expected

    @staticmethod
    def from_json(json_data):
        pbdata = qrl_pb2.StakeValidator()
        Parse(json_data, pbdata)
        return StakeValidator(pbdata)

    def to_json(self):
        return MessageToJson(self._data)
